import collections
import logging
from threading import Thread
from copy import copy, deepcopy
from collections import deque

logger = logging.getLogger(__name__)

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict


class DAGValidationError(Exception):
    pass


class DAG(object):
    """ Directed acyclic graph implementation. """

    def __init__(self):
        """ Construct a new DAG with no nodes or edges. """
        self.reset_graph()

    def add_node(self, node_name):
        """ Add a node if it does not exist yet, or error out.

        Args:
            node_name (str): The unique name of the node to add.

        Raises:
            KeyError: Raised if a node with the same name already exist in the
                      graph
        """
        graph = self.graph
        if node_name in graph:
            raise KeyError('node %s already exists' % node_name)
        graph[node_name] = set()

    def add_node_if_not_exists(self, node_name):
        """ Add a node if it does not exist yet, ignoring duplicates.

        Args:
            node_name (str): The name of the node to add.
        """
        try:
            self.add_node(node_name)
        except KeyError:
            pass

    def delete_node(self, node_name):
        """ Deletes this node and all edges referencing it.

        Args:
            node_name (str): The name of the node to delete.

        Raises:
            KeyError: Raised if the node does not exist in the graph.
        """
        graph = self.graph
        if node_name not in graph:
            raise KeyError('node %s does not exist' % node_name)
        graph.pop(node_name)

        for node, edges in graph.iteritems():
            if node_name in edges:
                edges.remove(node_name)

    def delete_node_if_exists(self, node_name):
        """ Deletes this node and all edges referencing it.

        Ignores any node that is not in the graph, rather than throwing an
        exception.

        Args:
            node_name (str): The name of the node to delete.
        """

        try:
            self.delete_node(node_name)
        except KeyError:
            pass

    def add_edge(self, ind_node, dep_node):
        """ Add an edge (dependency) between the specified nodes.

        Args:
            ind_node (str): The independent node to add an edge to.
            dep_node (str): The dependent node that has a dependency on the
                            ind_node.

        Raises:
            KeyError: Either the ind_node, or dep_node do not exist.
            DAGValidationError: Raised if the resulting graph is invalid.
        """
        graph = self.graph
        if ind_node not in graph:
            raise KeyError('independent node %s does not exist' % ind_node)
        if dep_node not in graph:
            raise KeyError('dependent node %s does not exist' % dep_node)
        test_graph = deepcopy(graph)
        test_graph[ind_node].add(dep_node)
        test_dag = DAG()
        test_dag.graph = test_graph
        is_valid, message = test_dag.validate()
        if is_valid:
            graph[ind_node].add(dep_node)
        else:
            raise DAGValidationError(message)

    def delete_edge(self, ind_node, dep_node):
        """ Delete an edge from the graph.

        Args:
            ind_node (str): The independent node to delete an edge from.
            dep_node (str): The dependent node that has a dependency on the
                            ind_node.

        Raises:
            KeyError: Raised when the edge doesn't already exist.
        """
        graph = self.graph
        if dep_node not in graph.get(ind_node, []):
            raise KeyError("No edge exists between %s and %s." % (
                ind_node, dep_node
                )
            )
        graph[ind_node].remove(dep_node)

    def transpose(self):
        """ Builds a new graph with the edges reversed.

        Returns:
            :class:`stacker.dag.DAG`: The transposed graph.
        """
        graph = self.graph
        transposed = DAG()
        for node, edges in graph.items():
            transposed.add_node(node)
        for node, edges in graph.items():
            # for each edge A -> B, transpose it so that B -> A
            for edge in edges:
                transposed.add_edge(edge, node)
        return transposed

    def walk(self, walk_func):
        """ Walks each node of the graph in reverse topological order.
        This can be used to perform a set of operations, where the next
        operation depends on the previous operation. It's important to note
        that walking happens serially, and is not paralellized.

        Args:
            walk_func (:class:`types.FunctionType`): The function to be called
                on each node of the graph.
        """
        nodes = self.topological_sort()
        # Reverse so we start with nodes that have no dependencies.
        nodes.reverse()

        for n in nodes:
            walk_func(n)

    def transitive_reduction(self):
        """ Performs a transitive reduction on the DAG. The transitive
        reduction of a graph is a graph with as few edges as possible with the
        same reachability as the original graph.

        See https://en.wikipedia.org/wiki/Transitive_reduction
        """

        graph = self.graph
        for x in graph.keys():
            for y in graph.keys():
                for z in graph.keys():
                    # Edge from x -> y
                    xy = y in graph[x]
                    # Edge from y -> x
                    yz = z in graph[y]
                    # Edge from x -> z
                    xz = z in graph[x]

                    # If edges xy and yz exist, remove edge xz.
                    if xz and xy and yz:
                        logger.debug("Removing edge %s -> %s" % (x, z))
                        graph[x].remove(z)

    def rename_edges(self, old_node_name, new_node_name):
        """ Change references to a node in existing edges.

        Args:
            old_node_name (str): The old name for the node.
            new_node_name (str): The new name for the node.
        """
        graph = self.graph
        for node, edges in graph.items():
            if node == old_node_name:
                graph[new_node_name] = copy(edges)
                del graph[old_node_name]

            else:
                if old_node_name in edges:
                    edges.remove(old_node_name)
                    edges.add(new_node_name)

    def predecessors(self, node):
        """ Returns a list of all immediate predecessors of the given node

        Args:
            node (str): The node whose predecessors you want to find.

        Returns:
            list: A list of nodes that are immediate predecessors to node.
        """
        graph = self.graph
        return [key for key in graph if node in graph[key]]

    def downstream(self, node):
        """ Returns a list of all nodes this node has edges towards.

        Args:
            node (str): The node whose downstream nodes you want to find.

        Returns:
            list: A list of nodes that are immediately downstream from the
                  node.
        """
        graph = self.graph
        if node not in graph:
            raise KeyError('node %s is not in graph' % node)
        return list(graph[node])

    def all_downstreams(self, node):
        """Returns a list of all nodes ultimately downstream
        of the given node in the dependency graph, in
        topological order.

        Args:
             node (str): The node whose downstream nodes you want to find.

        Returns:
            list: A list of nodes that are downstream from the node.
        """
        nodes = [node]
        nodes_seen = set()
        i = 0
        while i < len(nodes):
            downstreams = self.downstream(nodes[i])
            for downstream_node in downstreams:
                if downstream_node not in nodes_seen:
                    nodes_seen.add(downstream_node)
                    nodes.append(downstream_node)
            i += 1
        return filter(
                lambda node: node
                in nodes_seen, self.topological_sort())

    def filter(self, nodes):
        """ Returns a new DAG with only the given nodes and their
        dependencies.

        Args:
            nodes (list): The nodes you are interested in.

        Returns:
            :class:`stacker.dag.DAG`: The filtered graph.
        """

        filtered_dag = DAG()

        # Add only the nodes we need.
        for node in nodes:
            filtered_dag.add_node_if_not_exists(node)
            for edge in self.all_downstreams(node):
                filtered_dag.add_node_if_not_exists(edge)

        # Now, rebuild the graph for each node that's present.
        for node, edges in self.graph.items():
            if node in filtered_dag.graph:
                filtered_dag.graph[node] = edges

        return filtered_dag

    def all_leaves(self):
        """ Return a list of all leaves (nodes with no downstreams)

        Returns:
            list: A list of all the nodes with no downstreams.
        """
        graph = self.graph
        return [key for key in graph if not graph[key]]

    def from_dict(self, graph_dict):
        """ Reset the graph and build it from the passed dictionary.

        The dictionary takes the form of {node_name: [directed edges]}

        Args:
            graph_dict (dict): The dictionary used to create the graph.

        Raises:
            TypeError: Raised if the value of items in the dict are not lists.
        """

        self.reset_graph()
        for new_node in graph_dict.iterkeys():
            self.add_node(new_node)
        for ind_node, dep_nodes in graph_dict.iteritems():
            if not isinstance(dep_nodes, collections.Iterable):
                raise TypeError('%s: dict values must be lists' % ind_node)
            for dep_node in dep_nodes:
                self.add_edge(ind_node, dep_node)

    def reset_graph(self):
        """ Restore the graph to an empty state. """
        self.graph = OrderedDict()

    def ind_nodes(self):
        """ Returns a list of all nodes in the graph with no dependencies.

        Returns:
            list: A list of all independent nodes.
        """
        graph = self.graph

        dependent_nodes = set(
            node for dependents
            in graph.itervalues() for node in dependents)
        return [node for node in graph.keys() if node not in dependent_nodes]

    def validate(self):
        """ Returns (Boolean, message) of whether DAG is valid. """
        if len(self.ind_nodes()) == 0:
            return (False, 'no independent nodes detected')
        try:
            self.topological_sort()
        except ValueError as e:
            return (False, e.message)
        return (True, 'valid')

    def topological_sort(self):
        """ Returns a topological ordering of the DAG.

        Returns:
            list: A list of topologically sorted nodes in the graph.

        Raises:
            ValueError: Raised if the graph is not acyclic.
        """
        graph = self.graph

        in_degree = {}
        for u in graph:
            in_degree[u] = 0

        for u in graph:
            for v in graph[u]:
                in_degree[v] += 1

        queue = deque()
        for u in in_degree:
            if in_degree[u] == 0:
                queue.appendleft(u)

        sorted_graph = []
        while queue:
            u = queue.pop()
            sorted_graph.append(u)
            for v in graph[u]:
                in_degree[v] -= 1
                if in_degree[v] == 0:
                    queue.appendleft(v)

        if len(sorted_graph) == len(graph):
            return sorted_graph
        else:
            raise ValueError('graph is not acyclic')

    def size(self):
        return len(self)

    def __len__(self):
        return len(self.graph)


def walk(dag, walk_func):
    return dag.walk(walk_func)


class UnlimitedSemaphore(object):
    """UnlimitedSemaphore implements the same interface as threading.Semaphore,
    but acquire's always succeed.
    """

    def acquire(self, *args):
        pass

    def release(self):
        pass


class ThreadedWalker(object):
    """A DAG walker that walks the graph as quickly as the graph topology
    allows, using threads.

    Args:
        semaphore (threading.Semaphore, optional): a semaphore object which
            can be used to control how many steps are executed in parallel.
            By default, there is not limit to the amount of parallelism,
            other than what the graph topology allows.
    """

    def __init__(self, semaphore=None):
        self.semaphore = semaphore or UnlimitedSemaphore()

    def walk(self, dag, walk_func):
        """ Walks each node of the graph, in parallel if it can.
        The walk_func is only called when the nodes dependencies have been
        satisfied
        """

        # First, we'll topologically sort all of the nodes, with nodes that
        # have no dependencies first. We do this to ensure that we don't call
        # .join on a thread that hasn't yet been started.
        #
        # TODO(ejholmes): An alternative would be to ensure that Thread.join
        # blocks if the thread has not yet been started.
        nodes = dag.topological_sort()
        nodes.reverse()

        # This maps a node name to a thread of execution.
        threads = {}

        # Blocks until all of the given nodes have completed execution (whether
        # successfully, or errored). Returns True if all nodes returned True.
        def wait_for(nodes):
            for node in nodes:
                thread = threads[node]
                while thread.is_alive():
                    threads[node].join(0.5)

        # For each node in the graph, we're going to allocate a thread to
        # execute. The thread will block executing walk_func, until all of the
        # nodes dependencies have executed.
        for node in nodes:
            def fn(n, deps):
                if deps:
                    logger.debug(
                        "%s waiting for %s to complete",
                        n,
                        ", ".join(deps))

                # Wait for all dependencies to complete.
                wait_for(deps)

                logger.debug("%s starting", n)

                self.semaphore.acquire()
                try:
                    return walk_func(n)
                finally:
                    self.semaphore.release()

            deps = dag.all_downstreams(node)
            threads[node] = Thread(target=fn, args=(node, deps), name=node)

        # Start up all of the threads.
        for node in nodes:
            threads[node].start()

        # Wait for all threads to complete executing.
        wait_for(nodes)



"""Prints the the relationships between steps as a graph.

"""

from .base import BaseCommand
from ...actions import graph


class Graph(BaseCommand):

    name = "graph"
    description = __doc__

    def add_arguments(self, parser):
        super(Graph, self).add_arguments(parser)
        parser.add_argument("-f", "--format", default="dot",
                            choices=list(graph.FORMATTERS.iterkeys()),
                            help="The format to print the graph in.")
        parser.add_argument("--reduce", action="store_true",
                            help="When provided, this will create a "
                                 "graph with less edges, by performing "
                                 "a transitive reduction on the underlying "
                                 "graph. While this will produce a less "
                                 "noisy graph, it is slower.")

    def run(self, options, **kwargs):
        super(Graph, self).run(options, **kwargs)
        action = graph.Action(options.context,
                              provider_builder=options.provider_builder)
        action.execute(
            format=options.format,
            reduce=options.reduce)



import logging

from .base import BaseAction, plan, build_walker
from .base import STACK_POLL_TIME
from ..exceptions import StackDoesNotExist
from .. import util
from ..status import (
    CompleteStatus,
    SubmittedStatus,
    SUBMITTED,
    INTERRUPTED
)

from ..status import StackDoesNotExist as StackDoesNotExistStatus

logger = logging.getLogger(__name__)

DestroyedStatus = CompleteStatus("stack destroyed")
DestroyingStatus = SubmittedStatus("submitted for destruction")


class Action(BaseAction):
    """Responsible for destroying CloudFormation stacks.

    Generates a destruction plan based on stack dependencies. Stack
    dependencies are reversed from the build action. For example, if a Stack B
    requires Stack A during build, during destroy Stack A requires Stack B be
    destroyed first.

    The plan defaults to printing an outline of what will be destroyed. If
    forced to execute, each stack will get destroyed in order.

    """

    def _generate_plan(self, tail=False):
        return plan(
            description="Destroy stacks",
            action=self._destroy_stack,
            tail=self._tail_stack if tail else None,
            stacks=self.context.get_stacks(),
            targets=self.context.stack_names,
            reverse=True)

    def _destroy_stack(self, stack, **kwargs):
        old_status = kwargs.get("status")
        wait_time = STACK_POLL_TIME if old_status == SUBMITTED else 0
        if self.cancel.wait(wait_time):
            return INTERRUPTED

        provider = self.build_provider(stack)

        try:
            provider_stack = provider.get_stack(stack.fqn)
        except StackDoesNotExist:
            logger.debug("Stack %s does not exist.", stack.fqn)
            # Once the stack has been destroyed, it doesn't exist. If the
            # status of the step was SUBMITTED, we know we just deleted it,
            # otherwise it should be skipped
            if kwargs.get("status", None) == SUBMITTED:
                return DestroyedStatus
            else:
                return StackDoesNotExistStatus()

        logger.debug(
            "Stack %s provider status: %s",
            provider.get_stack_name(provider_stack),
            provider.get_stack_status(provider_stack),
        )
        if provider.is_stack_destroyed(provider_stack):
            return DestroyedStatus
        elif provider.is_stack_in_progress(provider_stack):
            return DestroyingStatus
        else:
            logger.debug("Destroying stack: %s", stack.fqn)
            provider.destroy_stack(provider_stack)
        return DestroyingStatus

    def pre_run(self, outline=False, *args, **kwargs):
        """Any steps that need to be taken prior to running the action."""
        pre_destroy = self.context.config.pre_destroy
        if not outline and pre_destroy:
            util.handle_hooks(
                stage="pre_destroy",
                hooks=pre_destroy,
                provider=self.provider,
                context=self.context)

    def run(self, force, concurrency=0, tail=False, *args, **kwargs):
        plan = self._generate_plan(tail=tail)
        if force:
            # need to generate a new plan to log since the outline sets the
            # steps to COMPLETE in order to log them
            plan.outline(logging.DEBUG)
            walker = build_walker(concurrency)
            plan.execute(walker)
        else:
            plan.outline(message="To execute this plan, run with \"--force\" "
                                 "flag.")

    def post_run(self, outline=False, *args, **kwargs):
        """Any steps that need to be taken after running the action."""
        post_destroy = self.context.config.post_destroy
        if not outline and post_destroy:
            util.handle_hooks(
                stage="post_destroy",
                hooks=post_destroy,
                provider=self.provider,
                context=self.context)



from collections import namedtuple

TYPE_NAME = "output"

Output = namedtuple("Output", ("stack_name", "output_name"))


def handler(value, context=None, **kwargs):
    """Fetch an output from the designated stack.

    Args:
        value (str): string with the following format:
            <stack_name>::<output_name>, ie. some-stack::SomeOutput
        context (:class:`stacker.context.Context`): stacker context

    Returns:
        str: output from the specified stack

    """

    if context is None:
        raise ValueError('Context is required')

    d = deconstruct(value)
    stack = context.get_stack(d.stack_name)
    return stack.outputs[d.output_name]


def deconstruct(value):

    try:
        stack_name, output_name = value.split("::")
    except ValueError:
        raise ValueError("output handler requires syntax "
                         "of <stack>::<output>.  Got: %s" % value)

    return Output(stack_name, output_name)



import operator


class Status(object):
    def __init__(self, name, code, reason=None):
        self.name = name
        self.code = code
        self.reason = reason or getattr(self, "reason", None)

    def _comparison(self, operator, other):
        if hasattr(other, "code"):
            return operator(self.code, other.code)
        return NotImplemented

    def __eq__(self, other):
        return self._comparison(operator.eq, other)

    def __ne__(self, other):
        return self._comparison(operator.ne, other)

    def __lt__(self, other):
        return self._comparison(operator.lt, other)

    def __gt__(self, other):
        return self._comparison(operator.gt, other)

    def __le__(self, other):
        return self._comparison(operator.le, other)

    def __ge__(self, other):
        return self._comparison(operator.ge, other)


class PendingStatus(Status):
    def __init__(self, reason=None):
        super(PendingStatus, self).__init__("pending", 0, reason)


class SubmittedStatus(Status):
    def __init__(self, reason=None):
        super(SubmittedStatus, self).__init__("submitted", 1, reason)


class CompleteStatus(Status):
    def __init__(self, reason=None):
        super(CompleteStatus, self).__init__("complete", 2, reason)


class SkippedStatus(Status):
    def __init__(self, reason=None):
        super(SkippedStatus, self).__init__("skipped", 3, reason)


class FailedStatus(Status):
    def __init__(self, reason=None):
        super(FailedStatus, self).__init__("failed", 4, reason)


class NotSubmittedStatus(SkippedStatus):
    reason = "disabled"


class NotUpdatedStatus(SkippedStatus):
    reason = "locked"


class DidNotChangeStatus(SkippedStatus):
    reason = "nochange"


class StackDoesNotExist(SkippedStatus):
    reason = "does not exist in cloudformation"


PENDING = PendingStatus()
SUBMITTED = SubmittedStatus()
COMPLETE = CompleteStatus()
SKIPPED = SkippedStatus()
FAILED = FailedStatus()
INTERRUPTED = FailedStatus(reason="interrupted")



import re
import base64

from ...util import read_value_from_path
from troposphere import GenericHelperFn, Base64

TYPE_NAME = "file"


def handler(value, **kwargs):
    """Translate a filename into the file contents.

    Fields should use the following format::

        <codec>:<path>

    For example::

        # We've written a file to /some/path:
        $ echo "hello there" > /some/path

        # In stacker we would reference the contents of this file with the
        # following
        conf_key: ${file plain:file://some/path}

        # The above would resolve to
        conf_key: hello there

        # Or, if we used wanted a base64 encoded copy of the file data
        conf_key: ${file base64:file://some/path}

        # The above would resolve to
        conf_key: aGVsbG8gdGhlcmUK

    Supported codecs:

        - plain

        - base64 - encode the plain text file at the given path with base64
          prior to returning it

        - parameterized - the same as plain, but additionally supports
          referencing template parameters to create userdata that's
          supplemented with information from the template, as is commonly
          needed in EC2 UserData. For example, given a template parameter of
          BucketName, the file could contain the following text::

            #!/bin/sh
            aws s3 sync s3://{{BucketName}}/somepath /somepath

          and then you could use something like this in the YAML config file::

            UserData: ${file parameterized:/path/to/file}

          resulting in the UserData parameter being defined as::

              { "Fn::Join" : ["", [
                  "#!/bin/sh\\naws s3 sync s3://",
                  {"Ref" : "BucketName"},
                  "/somepath /somepath"
              ]] }

        - parameterized-b64 - the same as parameterized, with the results
          additionally wrapped in *{ "Fn::Base64": ... }* , which is what you
          actually need for EC2 UserData

    When using parameterized-b64 for UserData, you should use a variable
    defined as such:

    .. code-block:: python

        from troposphere import AWSHelperFn

          "UserData": {
              "type": AWSHelperFn,
              "description": "Instance user data",
              "default": Ref("AWS::NoValue")
          }

    and then assign UserData in a LaunchConfiguration or Instance to
    *self.get_variables()["UserData"]*. Note that we use AWSHelperFn as the
    type because the parameterized-b64 codec returns either a Base64 or a
    GenericHelperFn troposphere object
    """

    try:
        codec, path = value.split(":", 1)
    except ValueError:
        raise TypeError(
            "File value must be of the format"
            " \"<codec>:<path>\" (got %s)" % (value)
        )

    value = read_value_from_path(path)

    return CODECS[codec](value)


def parameterized_codec(raw, b64):
    pattern = re.compile(r'{{([::|\w]+)}}')

    parts = []
    s_index = 0

    for match in pattern.finditer(raw):
        parts.append(raw[s_index:match.start()])
        parts.append({"Ref": match.group(1)})
        s_index = match.end()

    parts.append(raw[s_index:])
    result = {"Fn::Join": ["", parts]}

    # Note, since we want a raw JSON object (not a string) output in the
    # template, we wrap the result in GenericHelperFn (not needed if we're
    # using Base64)
    return Base64(result) if b64 else GenericHelperFn(result)


CODECS = {
    "plain": lambda x: x,
    "base64": base64.b64encode,
    "parameterized": lambda x: parameterized_codec(x, False),
    "parameterized-b64": lambda x: parameterized_codec(x, True)
}



"""The gitclean command."""
from .env import Env


class GitClean(Env):
    """Extend Env with execute to run the gitclean method."""

    def execute(self):
        """Clean directory of files."""
        self.gitclean()



"""The gen-sample command."""
from .base import Base


class GenSample(Base):
    """Extend Base with execute to run the module generators."""

    def execute(self):
        """Run selected module generator."""
        if self.options['cfn']:
            self.generate_sample_cfn_module()
        elif self.options['sls']:
            self.generate_sample_sls_module()
        elif self.options['stacker']:
            self.generate_sample_stacker_module()
        elif self.options['tf']:
            self.generate_sample_tf_module()



"""Gets information on the CloudFormation stacks based on the given config."""

from .base import BaseCommand
from ...actions import info


class Info(BaseCommand):

    name = "info"
    description = __doc__

    def add_arguments(self, parser):
        super(Info, self).add_arguments(parser)
        parser.add_argument("--stacks", action="append",
                            metavar="STACKNAME", type=str,
                            help="Only work on the stacks given. Can be "
                                 "specified more than once. If not specified "
                                 "then stacker will work on all stacks in the "
                                 "config file.")

    def run(self, options, **kwargs):
        super(Info, self).run(options, **kwargs)
        action = info.Action(options.context,
                             provider_builder=options.provider_builder)

        action.execute()

    def get_context_kwargs(self, options, **kwargs):
        return {"stack_names": options.stacks}



import logging

from .base import BaseAction
from .. import exceptions

logger = logging.getLogger(__name__)


class Action(BaseAction):
    """Get information on CloudFormation stacks.

    Displays the outputs for the set of CloudFormation stacks.

    """

    def run(self, *args, **kwargs):
        logger.info('Outputs for stacks: %s', self.context.get_fqn())
        for stack in self.context.get_stacks():
            provider = self.build_provider(stack)

            try:
                provider_stack = provider.get_stack(stack.fqn)
            except exceptions.StackDoesNotExist:
                logger.info('Stack "%s" does not exist.' % (stack.fqn,))
                continue

            logger.info('%s:', stack.fqn)
            if 'Outputs' in provider_stack:
                for output in provider_stack['Outputs']:
                    logger.info(
                        '\t%s: %s',
                        output['OutputKey'],
                        output['OutputValue']
                    )



# A lot of this code exists to deal w/ the broken ECS connect_to_region
# function, and will be removed once this pull request is accepted:
#   https://github.com/boto/boto/pull/3143
import logging

from stacker.session_cache import get_session

logger = logging.getLogger(__name__)


def create_clusters(provider, context, **kwargs):
    """Creates ECS clusters.

    Expects a "clusters" argument, which should contain a list of cluster
    names to create.

    Args:
        provider (:class:`stacker.providers.base.BaseProvider`): provider
            instance
        context (:class:`stacker.context.Context`): context instance

    Returns: boolean for whether or not the hook succeeded.

    """
    conn = get_session(provider.region).client('ecs')

    try:
        clusters = kwargs["clusters"]
    except KeyError:
        logger.error("setup_clusters hook missing \"clusters\" argument")
        return False

    if isinstance(clusters, basestring):
        clusters = [clusters]

    cluster_info = {}
    for cluster in clusters:
        logger.debug("Creating ECS cluster: %s", cluster)
        r = conn.create_cluster(clusterName=cluster)
        cluster_info[r["cluster"]["clusterName"]] = r
    return {"clusters": cluster_info}



"""The deploy command."""
from .env import Env


class Deploy(Env):
    """Extend Env with execute to run the deploy method."""

    def execute(self):
        """Run deployments."""
        self.deploy()



"""runway base module."""
from __future__ import print_function

from contextlib import contextmanager
from subprocess import check_call, check_output

import glob
import logging
import os
import platform
import shutil
import stat
import sys

import yaml

# from stacker.util import parse_cloudformation_template
# parse_cloudformation_template wraps yaml_parse; it would be better to call it
# from util but that would require sys.path shenanigans here
from ..embedded.stacker.awscli_yamlhelper import yaml_parse as parse_cloudformation_template  # noqa
from .. import __version__ as version

logging.basicConfig(level=logging.INFO)
LOGGER = logging.getLogger('runway')
EMBEDDED_LIB_PATH = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    'embedded'
)


class Base(object):  # noqa pylint: disable=too-many-instance-attributes,too-many-public-methods
    """Base class for deployer classes."""

    def __init__(self, options, env_vars=None, env_root=None,  # noqa pylint: disable=too-many-arguments
                 deploy_opts=None, module_root=None, runway_config_dir=None):
        """Initialize base class."""
        self.options = options

        if env_vars is None:
            self.env_vars = os.environ.copy()
        else:
            self.env_vars = env_vars

        if env_root is None:
            self.env_root = os.getcwd()
        else:
            self.env_root = env_root

        self.module_root = module_root

        if deploy_opts is None:
            self.deploy_opts = {}
        else:
            self.deploy_opts = deploy_opts

        if runway_config_dir is None:
            self.runway_config_path = os.path.join(
                self.env_root,
                'runway.yml'
            )
        else:
            self.runway_config_path = os.path.join(
                runway_config_dir,
                'runway.yml'
            )
        self._runway_config = None

        self.environment_override_name = 'DEPLOY_ENVIRONMENT'
        self._environment_name = None

    def update_env_vars(self, val):
        """Update env_vars dict with provided dict."""
        for key, value in list(val.items()):
            self.env_vars[key] = value

    def get_env_dirs(self):
        """Return list of directories in env_root."""
        repo_dirs = os.walk(self.env_root).next()[1]
        if '.git' in repo_dirs:
            repo_dirs.remove('.git')  # not relevant for any repo operations
        return repo_dirs

    def get_python_files_at_env_root(self):
        """Return list of python files in env_root."""
        return glob.glob(os.path.join(self.env_root, '*.py'))

    def get_yaml_files_at_env_root(self):
        """Return list of yaml files in env_root."""
        yaml_files = glob.glob(
            os.path.join(self.env_root, '*.yaml')
        )
        yml_files = glob.glob(
            os.path.join(self.env_root, '*.yml')
        )
        return yaml_files + yml_files

    def lint(self, base_dir=None, dirs_to_scan=None):
        """Call code linters."""
        from flake8.main import application as flake8_app
        from yamllint.cli import run as yamllint_run

        if base_dir is None:
            base_dir = self.env_root
        if dirs_to_scan is None:
            dirs_to_scan = self.get_env_dirs()

        if os.path.isfile(os.path.join(base_dir, '.flake8')):
            # config file in env will be picked up automatically
            flake8_config = []
        else:
            # no config file in env; use runway defaults
            flake8_config = [
                ('--append-config=' + os.path.join(
                    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),  # noqa
                    'templates',
                    '.flake8'
                ))
            ]
        if os.path.isfile(os.path.join(base_dir, '.yamllint.yml')):
            yamllint_config = os.path.join(base_dir, '.yamllint.yml')
        else:
            yamllint_config = os.path.join(
                os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
                'templates',
                '.yamllint.yml'
            )
        with self.change_dir(base_dir):
            with self.ignore_exit_code_0():
                LOGGER.info('Starting Flake8 linting...')
                flake8_run = flake8_app.Application()
                flake8_run.run(
                    flake8_config + dirs_to_scan +  self.get_python_files_at_env_root()  # noqa pylint: disable=line-too-long
                )
                flake8_run.exit()
            with self.ignore_exit_code_0():
                LOGGER.info('Flake8 linting complete.')
                LOGGER.info('Starting yamllint...')
                yamllint_run(
                    ["--config-file=%s" % yamllint_config] + dirs_to_scan + self.get_yaml_files_at_env_root()  # noqa pylint: disable=line-too-long
                )
            LOGGER.info('yamllint complete.')

    def get_cookbook_dirs(self, base_dir=None):
        """Find cookbook directories."""
        if base_dir is None:
            base_dir = self.env_root

        cookbook_dirs = []
        dirs_to_skip = set(['.git'])
        for root, dirs, files in os.walk(base_dir):  # pylint: disable=W0612
            dirs[:] = [d for d in dirs if d not in dirs_to_skip]
            for name in files:
                if name == 'metadata.rb':
                    if 'cookbook' in os.path.basename(os.path.dirname(root)):
                        cookbook_dirs.append(root)
        return cookbook_dirs

    def cookbook_tests(self, base_dir=None):
        """Run cookbook tests."""
        if base_dir is None:
            base_dir = self.env_root

        cookbook_dirs = self.get_cookbook_dirs(base_dir)

        if cookbook_dirs:
            if self.which('foodcritic') is None \
                    or self.which('cookstyle') is None:
                LOGGER.error('"foodcritic" and/or "cookstyle" not found -- '
                             'please ensure ChefDK is installed.')
                sys.exit(1)

            for path in cookbook_dirs:
                check_call(['foodcritic', '-f any', path])
                check_call(['cookstyle', '-P', path])

    def python_tests(self, base_dir=None, pylint_rc_file=None):  # noqa pylint: disable=too-many-branches,too-many-locals
        """Run python tests."""
        from pylint.lint import Run as PylintRun

        if base_dir is None:
            base_dir = self.env_root
        if pylint_rc_file is None:
            if os.path.isfile(os.path.join(base_dir, '.pylintrc')):
                pylint_rc_config = [
                    "--rcfile=%s" % os.path.join(base_dir, '.pylintrc')
                ]
            else:
                pylint_rc_config = []

        # Check all python files in repo
        dirs_to_skip = set(['.git',
                            'node_modules',
                            '.serverless'])
        for root, dirs, files in os.walk(base_dir):  # noqa pylint: disable=too-many-nested-blocks
            dirs[:] = [d for d in dirs if d not in dirs_to_skip]
            for name in files:
                if name[-3:] == ".py":
                    filepath = os.path.join(root, name)
                    LOGGER.info("Checking %s with pylint (\"No config file "
                                "found...\" messages can be ignored)",
                                filepath)
                    with self.use_embedded_pkgs():  # for embedded stacker
                        with self.ignore_exit_code_0():
                            PylintRun(pylint_rc_config + ['-E',  # ignore warn
                                                          filepath])
                    # Blueprints should output their template when executed
                    if (root.endswith('blueprints') and
                            not filepath.endswith('__init__.py')):
                        self.ensure_file_is_executable(filepath)
                        try:
                            shell_out_env = os.environ.copy()
                            if 'PYTHONPATH' in shell_out_env:
                                shell_out_env['PYTHONPATH'] = (
                                    "%s:%s" % (EMBEDDED_LIB_PATH,
                                               shell_out_env['PYTHONPATH'])
                                )
                            else:
                                shell_out_env['PYTHONPATH'] = EMBEDDED_LIB_PATH
                            cfn_template = check_output(
                                [sys.executable, filepath],
                                env=shell_out_env
                            )
                            if cfn_template == '':
                                raise ValueError("Template output should not "
                                                 "be empty!")
                            parse_cloudformation_template(cfn_template)
                        except:  # noqa - Bare except fine in this context
                            print("Error while checking %s for valid "
                                  "YAML/JSON output" % filepath)
                            raise

    def test(self):
        """Execute tests."""
        self.lint()
        self.cookbook_tests()
        self.python_tests()

    def save_existing_iam_env_vars(self):
        """Backup IAM environment variables for later restoration."""
        for i in ['AWS_ACCESS_KEY_ID', 'AWS_SECRET_ACCESS_KEY',
                  'AWS_SESSION_TOKEN']:
            if i in self.env_vars:
                self.update_env_vars({'OLD_' + i: self.env_vars[i]})

    def restore_existing_iam_env_vars(self):
        """Restore backed up IAM environment variables."""
        for i in ['AWS_ACCESS_KEY_ID', 'AWS_SECRET_ACCESS_KEY',
                  'AWS_SESSION_TOKEN']:
            if 'OLD_' + i in self.env_vars:
                self.update_env_vars({i: self.env_vars['OLD_' + i]})
            elif i in self.env_vars:
                self.env_vars.pop(i)

    def assume_role(self, role_arn, session_name=None, duration_seconds=None,
                    region='us-east-1'):
        """Assume IAM role."""
        import boto3
        if session_name is None:
            session_name = 'runway'
        assume_role_opts = {'RoleArn': role_arn,
                            'RoleSessionName': session_name}
        if duration_seconds:
            assume_role_opts['DurationSeconds'] = int(duration_seconds)
        sts_client = boto3.client('sts', region_name=region)
        LOGGER.info("Assuming role %s...", role_arn)
        response = sts_client.assume_role(**assume_role_opts)
        self.update_env_vars(
            {'AWS_ACCESS_KEY_ID': response['Credentials']['AccessKeyId'],
             'AWS_SECRET_ACCESS_KEY': response['Credentials']['SecretAccessKey'],  # noqa
             'AWS_SESSION_TOKEN': response['Credentials']['SessionToken']}
        )

    def pre_deploy_assume_role(self, assume_role_config, region):
        """Assume role (prior to deployment)."""
        if isinstance(assume_role_config, dict):
            if assume_role_config.get('post_deploy_env_revert'):
                self.save_existing_iam_env_vars()
            if assume_role_config.get('arn'):
                assume_role_arn = assume_role_config['arn']
                assume_role_duration = assume_role_config.get('duration')
            elif assume_role_config.get(self.environment_name):
                if isinstance(assume_role_config[self.environment_name], dict):
                    assume_role_arn = assume_role_config[self.environment_name]['arn']  # noqa
                    assume_role_duration = assume_role_config[self.environment_name].get('duration')  # noqa pylint: disable=line-too-long
                else:
                    assume_role_arn = assume_role_config[self.environment_name]
                    assume_role_duration = None
            else:
                LOGGER.info('Skipping assume-role; no role found for '
                            'environment %s...',
                            self.environment_name)
                return True

            self.assume_role(
                role_arn=assume_role_arn,
                session_name=assume_role_config.get('session_name', None),
                duration_seconds=assume_role_duration,
                region=region
            )
            return True
        else:
            self.assume_role(role_arn=assume_role_config, region=region)
            return True

    def post_deploy_assume_role(self, assume_role_config):
        """Assume role (prior to deployment)."""
        if isinstance(assume_role_config, dict):
            if assume_role_config.get('post_deploy_env_revert'):
                self.restore_existing_iam_env_vars()

    def path_only_contains_dirs(self, path):
        """Return boolean on whether a path only contains directories."""
        pathlistdir = os.listdir(path)
        if pathlistdir == []:
            return True
        elif any(os.path.isfile(os.path.join(path, i)) for i in pathlistdir):
            return False
        return all(self.path_only_contains_dirs(os.path.join(path, i)) for i in pathlistdir)  # noqa

    def get_empty_dirs(self, path):
        """Return a list of empty directories in path."""
        empty_dirs = []
        for i in os.listdir(path):
            child_path = os.path.join(path, i)
            if i == '.git' or os.path.isfile(child_path) or os.path.islink(child_path):  # noqa
                continue
            if self.path_only_contains_dirs(child_path):
                empty_dirs.append(i)
        return empty_dirs

    def generate_sample_sls_module(self, module_dir=None):
        """Generate skeleton Serverless sample module."""
        if module_dir is None:
            module_dir = os.path.join(self.env_root, 'sampleapp.sls')
        self.generate_sample_module(module_dir)
        for i in ['config-dev-us-east-1.json', 'handler.py', 'package.json',
                  'serverless.yml']:
            shutil.copyfile(
                os.path.join(os.path.dirname(os.path.dirname(__file__)),
                             'templates',
                             'serverless',
                             i),
                os.path.join(module_dir, i),
            )
        LOGGER.info("Sample Serverless module created at %s",
                    module_dir)

    def generate_sample_cfn_module(self, module_dir=None):
        """Generate skeleton CloudFormation sample module."""
        if module_dir is None:
            module_dir = os.path.join(self.env_root, 'sampleapp.cfn')
        self.generate_sample_module(module_dir)
        for i in ['01-sampleapp.yaml', 'dev-us-east-1.env']:
            shutil.copyfile(
                os.path.join(os.path.dirname(os.path.dirname(__file__)),
                             'templates',
                             'cfn',
                             i),
                os.path.join(module_dir, i)
            )
        os.mkdir(os.path.join(module_dir, 'templates'))
        shutil.copyfile(
            os.path.join(os.path.dirname(os.path.dirname(__file__)),
                         'templates',
                         'cfn',
                         'templates',
                         'bucket.json'),
            os.path.join(module_dir, 'templates', 'bucket.json')
        )
        LOGGER.info("Sample CloudFormation module created at %s",
                    module_dir)

    def generate_sample_stacker_module(self, module_dir=None):
        """Generate skeleton Stacker sample module."""
        if module_dir is None:
            module_dir = os.path.join(self.env_root, 'sampleapp.cfn')
        self.generate_sample_module(module_dir)
        for i in ['01-sampleapp.yaml', 'dev-us-east-1.env']:
            shutil.copyfile(
                os.path.join(os.path.dirname(os.path.dirname(__file__)),
                             'templates',
                             'stacker',
                             i),
                os.path.join(module_dir, i)
            )
        for i in ['sampleapp_blueprints', 'templates']:
            os.mkdir(os.path.join(module_dir, i))
        for i in list({'sampleapp_blueprints': ['__init__.py', 'bucket.py'],
                       'templates': ['bucket.json']}.items()):
            for template in i[1]:
                shutil.copyfile(
                    os.path.join(os.path.dirname(os.path.dirname(__file__)),
                                 'templates',
                                 'stacker',
                                 i[0],
                                 template),
                    os.path.join(module_dir, i[0], template)
                )
        os.chmod(  # make blueprint executable
            os.path.join(module_dir, 'sampleapp_blueprints', 'bucket.py'),
            os.stat(os.path.join(module_dir,
                                 'sampleapp_blueprints',
                                 'bucket.py')).st_mode | 0111
        )
        LOGGER.info("Sample Stacker module created at %s",
                    module_dir)

    def generate_sample_tf_module(self, module_dir=None):
        """Generate skeleton Terraform sample module."""
        if module_dir is None:
            module_dir = os.path.join(self.env_root, 'sampleapp.tf')
        self.generate_sample_module(module_dir)
        for i in ['.terraform-version', 'backend-us-east-1.tfvars',
                  'dev-us-east-1.tfvars', 'main.tf']:
            shutil.copyfile(
                os.path.join(os.path.dirname(os.path.dirname(__file__)),
                             'templates',
                             'terraform',
                             i),
                os.path.join(module_dir, i),
            )
        LOGGER.info("Sample Terraform app created at %s",
                    module_dir)

    def get_env(self, directory=None):
        """Determine environment name."""
        if self.environment_override_name in self.env_vars:
            return self.env_vars[self.environment_override_name]

        if self.runway_config.get('ignore_git_branch', False):
            LOGGER.info('Skipping environment lookup from current git branch '
                        '("ignore_git_branch" is set to true in the runway '
                        'config)')
        else:
            # These are not located with the top imports because they throw an
            # error if git isn't installed
            from git import Repo as GitRepo
            from git.exc import InvalidGitRepositoryError

            if directory is None:
                directory = self.module_root
            try:
                b_name = GitRepo(
                    directory,
                    search_parent_directories=True
                ).active_branch.name
                LOGGER.info('Deriving environment name from git branch %s...',
                            b_name)
                return self.get_env_from_branch(b_name)
            except InvalidGitRepositoryError:
                pass
        LOGGER.info('Deriving environment name from directory %s...',
                    self.env_root)
        return self.get_env_from_directory(os.path.basename(self.env_root))

    def parse_runway_config(self):
        """Read and parse runway.yml."""
        if not os.path.isfile(self.runway_config_path):
            LOGGER.error("Runway config file was not found (looking for "
                         "%s)",
                         self.runway_config_path)
            sys.exit(1)
        with open(self.runway_config_path) as data_file:
            return yaml.safe_load(data_file)

    def execute(self):
        """Implement dummy method (set in consuming classes)."""
        raise NotImplementedError('You must implement the execute() method '
                                  'yourself!')

    @property
    def environment_name(self):
        """Return environment name."""
        if not self._environment_name:
            self._environment_name = self.get_env()
        return self._environment_name

    @property
    def runway_config(self):
        """Return parsed runway.yml."""
        if not self._runway_config:
            self._runway_config = self.parse_runway_config()
        return self._runway_config

    @staticmethod
    def get_env_from_branch(branch_name):
        """Determine environment name from git branch name."""
        if branch_name.startswith('ENV-'):
            return branch_name[4:]
        elif branch_name == 'master':
            LOGGER.info('Translating git branch "master" to environment '
                        '"common"')
            return 'common'
        return branch_name

    @staticmethod
    def get_env_from_directory(directory_name):
        """Determine environment name from directory name."""
        if directory_name.startswith('ENV-'):
            return directory_name[4:]
        return directory_name

    @staticmethod
    def version():
        """Show current package version."""
        print(version)

    @staticmethod
    def which(program):
        """Mimic 'which' command behavior.

        Adapted from https://stackoverflow.com/a/377028
        """
        def is_exe(fpath):
            """Determine if program exists and is executable."""
            return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

        fpath, _fname = os.path.split(program)
        if fpath:
            if is_exe(program):
                return program
        else:
            for path in os.environ["PATH"].split(os.pathsep):
                exe_file = os.path.join(path, program)
                if is_exe(exe_file):
                    return exe_file

        return None

    @staticmethod
    def generate_sample_module(module_dir):
        """Generate skeleton sample module."""
        if os.path.isdir(module_dir):
            LOGGER.error("Error generating sample module -- directory %s "
                         "already exists!",
                         module_dir)
            sys.exit(1)
        os.mkdir(module_dir)

    @staticmethod
    def ensure_file_is_executable(path):
        """Exit if file is not executable."""
        if platform.system() != 'Windows' and (
                not stat.S_IXUSR & os.stat(path)[stat.ST_MODE]):
            print("Error: File %s is not executable" % path)
            sys.exit(1)

    @staticmethod
    @contextmanager
    def change_dir(newdir):
        """Change directory.

        Adapted from http://stackoverflow.com/a/24176022

        """
        prevdir = os.getcwd()
        os.chdir(os.path.expanduser(newdir))
        try:
            yield
        finally:
            os.chdir(prevdir)

    @staticmethod
    @contextmanager
    def override_sysargv(new_args=None):
        """Temporarily spoof the arguments provided in calling the application.

        Some applications will only detect their configuration from the command
        line. A list provided to this method will override what they detect.

        """
        if new_args is None:
            new_args = []
        orig_args = sys.argv
        sys.argv = new_args
        try:
            yield
        finally:
            sys.argv = orig_args

    @staticmethod
    @contextmanager
    def turn_down_stacker_logging(command):
        """Disable duplicate Stacker logging."""
        stacker_loggers = [
            'runway.embedded.stacker.actions.diff',
            'runway.embedded.stacker.commands.stacker',
            'runway.embedded.stacker.plan'
        ]
        if command == 'diff':
            try:
                for i in stacker_loggers:
                    logging.getLogger(i).setLevel(logging.ERROR)
                yield
            finally:
                for i in stacker_loggers:
                    logging.getLogger(i).setLevel(logging.INFO)
        else:
            yield

    @staticmethod
    @contextmanager
    def use_embedded_pkgs():
        """Temporarily prepend embedded packages to sys.path."""
        old_sys_path = sys.path
        sys.path.insert(
            1,  # https://stackoverflow.com/a/10097543
            EMBEDDED_LIB_PATH
        )
        try:
            yield
        finally:
            sys.path = old_sys_path

    @staticmethod
    @contextmanager
    def ignore_exit_code_0():
        """Capture exit calls and ignore those with exit code 0."""
        try:
            yield
        except SystemExit as exit_exc:
            if exit_exc.code != 0:
                raise



from string import Template

from .exceptions import InvalidLookupCombination
from .lookups import (
    extract_lookups,
    resolve_lookups,
)

from exceptions import FailedVariableLookup


class LookupTemplate(Template):

    """A custom string template we use to replace lookup values"""
    idpattern = r'[_a-z][^\$\{\}]*'


def resolve_string(value, replacements):
    """Resolve any lookups within a string.

    Args:
        value (str): string value we're resolving lookups within
        replacements (dict): resolved lookup values

    Returns:
        str: value with any lookups resolved

    """
    lookups = extract_lookups(value)
    for lookup in lookups:
        lookup_value = replacements.get(lookup.raw)
        if not isinstance(lookup_value, basestring):
            if len(lookups) > 1:
                raise InvalidLookupCombination(lookup, lookups, value)
            return lookup_value
    # we use safe_substitute to support resolving nested lookups
    return LookupTemplate(value).safe_substitute(replacements)


def resolve(value, replacements):
    """Recursively resolve any lookups within the data structure.

    Args:
        value (Union[str, list, dict]): a structure that contains lookups
        replacements: resolved lookup values

    Returns:
        Union[str, list, dict]: value passed in with lookup values resolved

    """
    if isinstance(value, basestring):
        return resolve_string(value, replacements)
    elif isinstance(value, list):
        resolved = []
        for v in value:
            resolved.append(resolve(v, replacements))
        return resolved
    elif isinstance(value, dict):
        for key, v in value.iteritems():
            value[key] = resolve(v, replacements)
        return value
    return value


def resolve_variables(variables, context, provider):
    """Given a list of variables, resolve all of them.

    Args:
        variables (list of :class:`stacker.variables.Variable`): list of
            variables
        context (:class:`stacker.context.Context`): stacker context
        provider (:class:`stacker.provider.base.BaseProvider`): subclass of the
            base provider

    """
    for variable in variables:
        variable.resolve(context, provider)


class Variable(object):

    """Represents a variable passed to a stack.

    Args:
        name (str): Name of the variable
        value (str): Initial value of the variable from the config

    """

    def __init__(self, name, value):
        self.name = name
        self._value = value
        self._resolved_value = None

    @property
    def lookups(self):
        """Return any lookups within the value"""
        return extract_lookups(self.value)

    @property
    def needs_resolution(self):
        """Return True if the value has any lookups that need resolving."""
        if self.lookups:
            return True
        return False

    @property
    def value(self):
        """Return the current value of the Variable.

        `_resolved_value` takes precedence over `_value`.

        """
        if self._resolved_value is not None:
            return self._resolved_value
        else:
            return self._value

    @property
    def resolved(self):
        """Boolean for whether the Variable has been resolved.

        Variables only need to be resolved if they contain lookups.

        """
        if self.needs_resolution:
            return self._resolved_value is not None
        return True

    def resolve(self, context, provider):
        """Recursively resolve any lookups with the Variable.

        Args:
            context (:class:`stacker.context.Context`): Current context for
                building the stack
            provider (:class:`stacker.provider.base.BaseProvider`): subclass of
                the base provider

        """

        while self.lookups:
            try:
                resolved_lookups = resolve_lookups(
                    self.lookups, context, provider)
            except Exception as e:
                raise FailedVariableLookup(self.name, e)

            self.replace(resolved_lookups)

    def replace(self, resolved_lookups):
        """Replace lookups in the Variable with their resolved values.

        Args:
            resolved_lookups (dict): dict of :class:`stacker.lookups.Lookup` ->
                resolved value.

        """
        replacements = {}
        for lookup, value in resolved_lookups.iteritems():
            replacements[lookup.raw] = value

        self._resolved_value = resolve(self.value, replacements)



"""The show-env command."""
from __future__ import print_function

import logging
import os

from .base import Base


class WhichEnv(Base):
    """Extend Base with execute to run the get_env method."""

    def execute(self):
        """Output environment name."""
        # Disable other runway logging so the only response is the env name
        logging.getLogger('runway').setLevel(logging.ERROR)

        # This may be invoked from a module directory in an environment;
        # account for that here if necessary
        if not os.path.isfile('runway.yml'):
            self.runway_config_path = os.path.join(
                os.path.dirname(os.getcwd()),
                'runway.yml')

        print(self.get_env())



""" Diffs the config against the currently running CloudFormation stacks

Sometimes small changes can have big impacts.  Run "stacker diff" before
"stacker build" to detect bad things(tm) from happening in advance!
"""

from .base import BaseCommand
from ...actions import diff


class Diff(BaseCommand):
    name = "diff"
    description = __doc__

    def add_arguments(self, parser):
        super(Diff, self).add_arguments(parser)
        parser.add_argument("--force", action="append", default=[],
                            metavar="STACKNAME", type=str,
                            help="If a stackname is provided to --force, it "
                                 "will be diffed, even if it is locked in "
                                 "the config.")
        parser.add_argument("--stacks", action="append",
                            metavar="STACKNAME", type=str,
                            help="Only work on the stacks given. Can be "
                                 "specified more than once. If not specified "
                                 "then stacker will work on all stacks in the "
                                 "config file.")

    def run(self, options, **kwargs):
        super(Diff, self).run(options, **kwargs)
        action = diff.Action(options.context,
                             provider_builder=options.provider_builder)
        action.execute()

    def get_context_kwargs(self, options, **kwargs):
        return {"stack_names": options.stacks, "force_stacks": options.force}
